	; Small C Z80 for GameBoy.

	.include	"lib/global.h"

	; Main user routine
	.globl	_main

	; ****************************************
	; Beginning of module
	.title	"GB Runtime"
	.module	Runtime
	.area	HEADER	(ABS)

	; Standard header for the GB.
	.org	0x0

	; Set all irq vectors to do nothing.
	.org	0x40
	RETI

	.org	0x48
	RETI

	.org	0x50
	RETI

	.org	0x58
	RETI

	.org	0x60
	RETI

	; GameBoy Header

	.org	0x100
	; Standard Nintendo DO NOT CHANGE...
	; Next line is the first instruction: JP 0x150.
	.byte	0x00,0xC3,0x50,0x01
	.byte	0xCE,0xED,0x66,0x66
	.byte	0xCC,0x0D,0x00,0x0B
	.byte	0x03,0x73,0x00,0x83
	.byte	0x00,0x0C,0x00,0x0D
	.byte	0x00,0x08,0x11,0x1F
	.byte	0x88,0x89,0x00,0x0E
	.byte	0xDC,0xCC,0x6E,0xE6
	.byte	0xDD,0xDD,0xD9,0x99
	.byte	0xBB,0xBB,0x67,0x63
	.byte	0x6E,0x0E,0xEC,0xCC
	.byte	0xDD,0xDC,0x99,0x9F
	.byte	0xBB,0xB9,0x33,0x3E

	.org	0x134
	; Title of the game.
	.asciZ	"test"

	.org	0x144
	; Not used.

	.org	0x147
	; Cartridge type is ROM only.
	.byte	0

	.org	0x148
	; ROM siZe is 32kB.
	.byte	0

	.org	0x149
	; RAM siZe is 0kB.
	.byte	0

	.org	0x14A
	; Maker ID.
	.byte	0x00,0x00

	.org	0x14C
	; Version number.
	.byte	0x01

	.org	0x14D
	; Complement check.
	; Generate it with CRC.EXE.
	.byte	0x00

	.org	0x14E
	; Checksum.
	; Generate it with CRC.EXE.
	.byte	0x00,0x00

	; ****************************************
	.org	0x150
	; Beginning of the code.
	DI
	LD		SP,#0xFFF4		; Put the stack where the GB wants it

	; Call the initialization routine.
	CALL	.init

	; Call the main function.
	CALL	_main

1$:
	JR		1$				; Wait forever

	.org	.MODE_TABLE
	; Jump table for modes.
	RET

	; ****************************************

	; Ordering of segments for the linker
	.area	_CODE
	.area	_DATA

	; Runtime library
	.area	_CODE

	; Fetch char from (HL) and sign extend into HL
.gchar::
	LD		A,(HL)
.sxt::
	LD		L,A
	RLCA
	SBC		A
	LD		H,A
	RET

	; Fetch int from (HL)
.gint::
	LD		A,(HL)
	INC		HL
	LD		H,(HL)
	LD		L,A
	RET

	; Store char from HL into (DE)
.pchar::
	LD		A,L
	LD		(DE),A
	RET

	; Store int from HL into (DE)
.pint::
	LD		A,L
	LD		(DE),A
	INC		DE
	LD		A,H
	LD		(DE),A
	RET

	; "Or" HL and DE into HL
.or::
	LD		A,L
	OR		E
	LD		L,A
	LD		A,H
	OR		D
	LD		H,A
	RET

	; "Xor" HL and DE into HL
.xor::
	LD		A,L
	XOR		E
	LD		L,A
	LD		A,H
	XOR		D
	LD		H,A
	RET

	; "And" HL and DE into HL
.and::
	LD		A,L
	AND		E
	LD		L,A
	LD		A,H
	AND		D
	LD		H,A
	RET

	; Logical operations: HL set to 0 (false) OR 1 (true)

	; DE == HL
.eq::
	CALL	.cmp
	RET		Z
	DEC		HL
	RET

	; DE != HL
.ne::
	CALL	.cmp
	RET		NZ
	DEC		HL
	RET

	; DE > HL [signed]
.gt::
	PUSH	HL
	PUSH	DE
	POP		HL
	POP		DE
	CALL	.cmp
	RET		C
	DEC		HL
	RET

	; DE <= HL [signed]
.le::
	CALL	.cmp
	RET		Z
	RET		C
	DEC		HL
	RET

	; DE >= HL [signed]
.ge::
	CALL	.cmp
	RET		NC
	DEC		HL
	RET

	; DE < HL [signed]
.lt::
	CALL	.cmp
	RET		C
	DEC		HL
	RET

	; DE >= HL [unsigned]
.uge::
	CALL	.ucmp
	RET		NC
	DEC		HL
	RET

	; DE < HL [unsigned]
.ult::
	CALL	.ucmp
	RET		C
	DEC		HL
	RET

	; DE > HL [unsigned]
.ugt::
	PUSH	HL
	PUSH	DE
	POP		HL
	POP		DE
	CALL	.ucmp
	RET		C
	DEC		HL
	RET

	; DE <= HL [unsigned]
.ule::
	CALL	.ucmp
	RET		Z
	RET		C
	DEC		HL
	RET

	; Signed compare of DE and HL
	; Carry is sign of difference [set => DE < HL]
	; Zero is Zero/non-Zero
.cmp::
	LD		A,E
	SUB		L
	LD		E,A
	LD		A,D
	SBC		H
	LD		HL,#0x01		; Preset true
	CP		#0x80
	JR		NC,1$
	OR		E				; Update Z and reset carry
	RET
1$:
	OR		E				; Update Z
	SCF						; Set carry
	RET

	; Unsigned compare of DE and HL
	; Carry is sign of difference [set => DE < HL]
	; Zero is Zero/non-Zero
.ucmp::
	LD		A,D
	CP		H
	JR		NZ,1$
	LD		A,E
	CP		L
1$:
	LD		HL,#0x01		; Preset true
	RET

	; Shift DE right arithmetically by HL, move to HL
.asr::
	PUSH	HL
	PUSH	DE
	POP		HL
	POP		DE
1$:
	DEC		E
	LD		A,E
	CP		#0x80
	RET		NC
	RL		H				; Store sign bit in C
	RR		H				; Rotate right
	RR		L
	JR		1$

	; Shift DE left arithmetically by HL, move to HL
.asl::
	PUSH	HL
	PUSH	DE
	POP		HL
	POP		DE
1$:
	DEC		E
	LD		A,E
	CP		#0x80
	RET		NC
	ADD		HL,HL
	JR		1$

	; HL = DE - HL
.sub::
	LD		A,E
	SUB		L
	LD		L,A
	LD		A,D
	SBC		H
	LD		H,A
	RET

	; HL = -HL
.neg::
	CALL	.com
	INC		HL
	RET

	; HL = ~HL
.com::
	LD		A,H
	CPL
	LD		H,A
	LD		A,L
	CPL
	LD		L,A
	RET

	; HL = !HL
.lneg::
	LD		A,H
	OR		L
	JR		Z,1$
	LD		HL,#0x00
	RET
1$:
	INC		HL
	RET

	; HL = !!HL
.bool::
	CALL	.lneg
	JP		.lneg

	; HL = DE * HL [signed]
.mul::
	LD		B,H
	LD		C,L
	LD		HL,#0x00
1$:
	LD		A,C
	RRCA
	JR		NC,2$
	ADD		HL,DE
2$:
	XOR		A				; reset carry
	RR		B
	RR		C
	LD		A,C
	OR		B
	RET		Z
	XOR		A				; Reset carry
	RL		E
	RL		D
	LD		A,D
	OR		E
	RET		Z
	JR		1$

	; HL = DE / HL, DE = DE % HL
.div::
	LD		B,H
	LD		C,L
	LD		A,D
	XOR		B
	CP		#0x80			; Update carry before pushing
	PUSH	AF
	LD		A,D
	CP		#0x80
	CALL	NC,4$
	LD		A,B
	CP		#0x80
	CALL	NC,5$
	LD		A,#0x10
	CP		#0x80			; Update carry before pushing
	PUSH	AF
	PUSH	DE
	POP		HL
	LD		DE,#0x00
1$:
	ADD		HL,HL
	CALL	6$
	JR		Z,2$
	CALL	7$
	CP		#0x80
	JR		NC,2$
	LD		A,L
	OR		#0x01
	LD		L,A
	LD		A,E
	SUB		C
	LD		E,A
	LD		A,D
	SBC		B
	LD		D,A
2$:
	POP		AF
	DEC		A
	JR		Z,3$
	CP		#0x80			; Update carry before pushing
	PUSH	AF
	JR		1$
3$:
	POP		AF
	RET		C				; Carry was set before pushing
	CALL	4$
	PUSH	HL
	PUSH	DE
	POP		HL
	POP		DE
	CALL	4$
	PUSH	HL
	PUSH	DE
	POP		HL
	POP		DE
	RET

	; DE = -DE
4$:
	LD		A,D
	CPL
	LD		D,A
	LD		A,E
	CPL
	LD		E,A
	INC		DE
	RET

	; BC = -BC
5$:
	LD		A,B
	CPL
	LD		B,A
	LD		A,C
	CPL
	LD		C,A
	INC		BC
	RET

	; DE << 1
6$:
	RL		E
	RL		D
	LD		A,D
	OR		E
	RET

	; BC : DE
7$:
	LD		A,E
	SUB		C
	LD		A,D
	SBC		B
	RET

	; Case jump
.case::
	PUSH	HL				; Switch value to DE
	POP		DE
	POP		HL				; Get table address
1$:
	CALL	4$				; Get case value
	LD		A,E
	CP		C				; Equal to switch value ?
	JR		NZ,2$			; No
	LD		A,D
	CP		B				; Equal to switch value ?
	JR		NZ,2$			; No
	CALL	4$				; Get case label
	JR		Z,3$			; End of table, go to default
	PUSH	BC
	RET						; Case jump
2$:
	CALL	4$				; Get case label
	JR		NZ,1$			; Next case
3$:
	DEC		HL
	DEC		HL
	DEC		HL
	LD		D,(HL)
	DEC		HL
	LD		E,(HL)
	PUSH	HL
	PUSH	DE
	POP		HL
	POP		DE
	JP		(HL)			; Default jump
4$:
	LD		C,(HL)
	INC		HL
	LD		B,(HL)
	INC		HL
	LD		A,C
	OR		B
	RET

	.area	_BSS

.mode:
	.ds		0x01

	.area	_CODE

.init:
	LD		A,#0x00			; No IRQs at all
	LDH		(0xFF),A
	LD		(.mode),A

	SUB		A				; Misc standard init things
	LDH		(.LCDSTAT),A	; LCDC Status
	LDH		(.SCROLLY),A	; Screen scroll Y=0
	LDH		(.SCROLLX),A	; Screen scroll X=0

	LD		A,#0b10000000
	LDH		(.LCDCONT),A

;	WindowBank	=	0x9800 (Not used)
;	Window		=	Off
;	BG Chr		=	0x8800
;	BG Bank		=	0x9800
;	OBJ			=	8x8
;	OBJ			=	Off
;	BG			=	Off

	RET

	; Call the initialization function for the mode specified in HL
.setmode:
	LD		A,L
	LD		(.mode),A

	LD		BC,#.MODE_TABLE
	RLC		L				; Multiply mode by 4
	RLC		L
	ADD		HL,BC
	JP		(HL)			; Jump to initialization routine

	; Wait until all buttons have been released and debounced
.padup:
	PUSH	AF				; Save modified registers
	PUSH	BC
1$:
	LD		B,#0xFF
2$:
	CALL	.jpad
	OR		A				; Have all buttons been released ?
	JR		NZ,1$			; Not yet

	DEC		B
	JR		NZ,2$
	POP		BC				; Restore registers
	POP		AF
	RET

	; Get Keypad Button Status
	; The following bits are set if pressed:
	;   0x80 - Start   0x08 - Down
	;   0x40 - Select  0x04 - Up
	;   0x20 - B       0x02 - Left
	;   0x10 - A       0x01 - Right
.jpad:
	PUSH	BC				; Save modified registers
	LD		A,#0x20
	LD		(0xFF00),A		; Turn on P15

	LD		A,(0xFF00)		; Delay
	LD		A,(0xFF00)
	CPL
	AND		#0x0F
	SWAP	A
	LD		B,A
	LD		A,#0x10
	LD		(0xFF00),A		; Turn on P14
	LD		A,(0xFF00)		; Delay
	LD		A,(0xFF00)
	LD		A,(0xFF00)
	LD		A,(0xFF00)
	LD		A,(0xFF00)
	LD		A,(0xFF00)
	CPL
	AND		#0x0F
	OR		B
	SWAP	A
	POP		BC				; Restore registers
	RET

	; Wait for the key in B to be pressed
.waitpad:
1$:
	CALL	.jpad			; Hold startup screen until start is pressed
	AND		B				; Is start pressed ?
	JR		Z,1$			; No
	RET

	; Wait for VBL
.waitvbl:
	LDH		A,(.LCDCONT)
	ADD		A,A
	RET		NC

1$:
	LDH		A,(.CURLINE)	; 0xFF44 = LCDC Y-Pos
	CP		#0x90			; 0x90 and bigger = in VBL
	JR		NZ,1$			; Loop until it $90
	RET

	; Wait for BC ns
.delay_ns:
1$:
	DEC		BC
	LD		A,B
	OR		C
	JR		NZ,1$

	RET

	; Wait for BC ms
.delay_ms:
	PUSH	DE
	LD		DE,#00
1$:
	DEC		DE
	LD		A,D
	OR		E
	JR		NZ,1$
	DEC		BC
	LD		A,B
	OR		C
	JR		NZ,1$
	POP		DE

	RET

	; All colors to transparent
.white:
	LD		A,#0x00
	LDH		(.BGRDPAL),A
	RET

	; All colors to black
.black:
	LD		A,#0xFF
	LDH		(.BGRDPAL),A
	RET

	; Sets the colors to normal palette
.std_pal:
	LD		A,#0b11100100	; Grey 3 = 11 (Black)
							; Grey 2 = 10 (Dark grey)
							; Grey 1 = 01 (Light grey)
							; Grey 0 = 00 (Transparent)
	LDH		(.BGRDPAL),A
	LDH		(.SPR0PAL),A
	LDH		(.SPR1PAL),A
	RET

	; Initialize the tile set with B
.init_tiles:
	LD		HL,#0x8000
	LD		DE,#0x0200

1$:
	LD		A,B
	LD		(HL+),A			; Unroll part of the loop
	LD		(HL+),A
	LD		(HL+),A
	LD		(HL+),A
	LD		(HL+),A
	LD		(HL+),A
	LD		(HL+),A
	LD		(HL+),A
	DEC		DE
	LD		A,D
	OR		E
	JR		NZ,1$
	RET

	; Initialize part of the tile set from (BC) to (HL)
	; of size DE bytes
.mv_tiles:
1$:
	LDH		A,(.LCDSTAT)
	AND		#0x02
	JR		NZ,1$

	LD		A,(BC)
	LD		(HL+),A
	INC		BC
	DEC		DE
	LD		A,D
	OR		E
	JR		NZ,1$
	RET

	; Set background tile table to table starting at (BC)
.set_btt:
	LD		HL,#0x9800
	JR		.set_tt

	; Set window tile table to table starting at (BC)
.set_wtt:
	LD		HL,#0x9C00

.set_tt:
	LD		DE,#0x0400		; One whole GB Screen

1$:
	LDH		A,(.LCDSTAT)
	AND		#0x02
	JR		NZ,1$

	LD		A,(BC)
	LD		(HL+),A
	INC		BC
	DEC		DE
	LD		A,D
	OR		E
	JR		NZ,1$

	RET

	; Set background tile table to B
.init_btt:
	LD		HL,#0x9800
	JR		.init_tt

	; Set window tile table to B
.init_wtt:
	LD		HL,#0x9C00

.init_tt:
	LD		DE,#0x0400		; One whole GB Screen

1$:
	LDH		A,(.LCDSTAT)
	AND		#0x02
	JR		NZ,1$

	LD		A,B
	LD		(HL+),A

	DEC		DE
	LD		A,D
	OR		E
	JR		NZ,1$

	RET

	; Set background tile table from BC at xy = DE of size wh = HL
	; wh >= (1,1)
.set_xy_btt:
	PUSH	HL
	LD		HL,#0x9800		; Calculate origin
	JR		.set_xy_tt

	; Set window tile table from BC at xy = DE of size wh = HL
	; wh >= (1,1)
.set_xy_wtt:
	PUSH	HL
	LD		HL,#0x9C00		; Calculate origin

.set_xy_tt:
	PUSH	BC
	LD		A,#0x00
	OR		E
	JR		Z,2$

	LD		BC,#0x20
1$:
	ADD		HL,BC			; y coordinate
	DEC		E
	JR		NZ,1$
2$:
	LD		B,#0x00			; x coordinate
	LD		C,D
	ADD		HL,BC

	POP		BC
	POP		DE				; DE = wh
	PUSH	HL				; Store origin
	PUSH	DE				; Store wh
3$:
	LDH		A,(.LCDSTAT)
	AND		#0x02
	JR		NZ,3$

	LD		A,(BC)
	LD		(HL+),A
	INC		BC
	DEC		D
	JR		NZ,3$
	POP		HL				; Get wh
	LD		D,H				; Restore D = w
	POP		HL				; Get origin
	DEC		E
	JR		Z,4$

	PUSH	BC				; Next line
	LD		BC,#0x20
	ADD		HL,BC
	POP		BC

	PUSH	HL				; Store current line
	PUSH	DE				; Store wh
	JR		3$
4$:
	RET

	; Move sprite number C at xy = DE
.mv_sprite:
	LD		HL,#0xFE00		; Calculate origin of sprite info

	RLC		C				; Multiply C by 4
	RLC		C
	LD		B,#0x00
	ADD		HL,BC

1$:
	LDH		A,(.LCDSTAT)
	AND		#0x02
	JR		NZ,1$

	LD		A,E
	LD		(HL+),A

2$:							; Have we the time to make both transfers ?
	LDH		A,(.LCDSTAT)
	AND		#0x02
	JR		NZ,2$

	LD		A,D
	LD		(HL+),A
	RET

	; Set sprite number C to tile D
.set_sprite:
	LD		HL,#0xFE00		; Calculate origin of sprite info

	RLC		C				; Multiply C by 4
	RLC		C
	LD		B,#0x00
	ADD		HL,BC
	INC		HL				; Add 2
	INC		HL

1$:
	LDH		A,(.LCDSTAT)
	AND		#0x02
	JR		NZ,1$

	LD		A,D
	LD		(HL),A
	RET

	; Set properties of sprite number C to D
.prop_sprite:
	LD		HL,#0xFE00		; Calculate origin of sprite info

	RLC		C				; Multiply C by 4
	RLC		C
	LD		B,#0x00
	ADD		HL,BC
	INC		HL				; Add 3
	INC		HL
	INC		HL

1$:
	LDH		A,(.LCDSTAT)
	AND		#0x02
	JR		NZ,1$

	LD		A,D
	LD		(HL),A
	RET
