#include <stdio.h>
#include <ctype.h>
#include "defs.h"
#include "data.h"
#include "decls.h"

int primary(lval_t *lval)
{
  char sname[NAMESIZE];
  sym_t *ptr;
  int num[1];
  int k;

  lval->type = 0;  /* clear pointer/array type */
  if (match ("(")) {
    k = heir1 (lval);
    needbrack (")");
    return (k);
  }
  if (amatch("sizeof", 6)) {
    needbrack("(");
    immed();
    if (amatch("int", 3)) onum(intsize());
    else if (amatch("char", 4)) onum(1);
    else if (symname(sname)) {
      if ((ptr = findloc(sname)) ||
	  (ptr = findglb(sname))) {
	if (ptr->storage == LSTATIC)
	  error("sizeof local static");
	k = glint(ptr);
	if ((ptr->type == CINT) ||
	    (ptr->ident == POINTER))
	  k *= intsize();
	onum(k);
      } else {
	error("sizeof undeclared variable");
	onum(0);
      }
    } else {
      error("sizeof only on type or variable");
    }
    needbrack(")");
    nl();
    lval->addr = 0;
    lval->what = 0;
    return(0);
  }
  if (symname (sname)) {
    if ((ptr = findloc (sname))) {
      getloc (ptr);
      lval->addr = ptr;
      lval->what =  ptr->type;
      if (ptr->ident == POINTER) {
	lval->what = CINT;
	lval->type = ptr->type;
      }
      if (ptr->ident == ARRAY) {
	lval->type = ptr->type;
	lval->type = 0;
	return (0);
      }
      else
	return (1);
    }
    if ((ptr = findglb (sname)))
      if (ptr->ident != FUNCTION) {
	lval->addr = ptr;
	lval->what = 0;
	if (ptr->ident != ARRAY) {
	  if (ptr->ident == POINTER)
	    lval->type = ptr->type;
	  return (1);
	}
	immed ();
	prefix ();
	outstr (ptr->name);
	nl ();
	lval->what = lval->type = ptr->type;
	lval->type = 0;
	return (0);
      }
    blanks ();
    if (ch() != '(')
      error("undeclared variable");
    ptr = addglb (sname, FUNCTION, CINT, 0, PUBLIC);
    lval->addr = ptr;
    lval->what = 0;
    return (0);
  }
  if (constant (num)) {
    lval->addr = 0;
    lval->what = 0;
    return(0);
  } else {
    error ("invalid expression");
    immed ();
    onum (0);
    nl ();
    junk ();
    return (0);
  }
}

/*
 *	true if val1 -> int pointer or int array and val2 not pointer or array
 */
int dbltest(lval_t *lval1, lval_t *lval2)
{
  if (lval1 == NULL)
    return (FALSE);
  if (lval1->type != CINT)
    return (FALSE);
  if (lval2->type)
    return (FALSE);
  return (TRUE);
}

/*
 *	determine type of binary operation
 */
void result(lval_t *lval, lval_t *lval2)
{
  if (lval->type && lval2->type)
    lval->type = 0;
  else if (lval2->type) {
    lval->addr = lval2->addr;
    lval->what = lval2->what;
    lval->type = lval2->type;
  }
}

int constant(int *val)
{
  if (number (val))
    immed ();
  else if (pstr (val))
    immed ();
  else if (qstr (val)) {
    immed ();
    printlabel (litlab);
    outbyte ('+');
  } else
    return (0);
  onum (val[0]);
  nl ();
  return (1);
}

int number(int *val)
{
  int k, minus, base;
  char c;

  k = minus = 1;
  while (k) {
    k = 0;
    if (match ("+"))
      k = 1;
    if (match ("-")) {
      minus = (-minus);
      k = 1;
    }
  }
  if (!isdigit (c = ch ()))
    return (0);
  if (match ("0x") || match ("0X"))
    while (isdigit (c = ch ()) ||
	   (c >= 'a' && c <= 'f') ||
	   (c >= 'A' && c <= 'F')) {
      inpbyte ();
      k = k * 16 +
	(isdigit (c) ? (c - '0') : ((c & 07) + 9));
    }
  else {
    base = (c == '0') ? 8 : 10;
    while (isdigit (ch ())) {
      c = inpbyte ();
      k = k * base + (c - '0');
    }
  }
  if (minus < 0)
    k = (-k);
  val[0] = k;
  return (1);
}

int pstr(int *val)
{
  int k;
  char c;

  k = 0;
  if (!match ("'"))
    return (0);
  while ((c = gch ()) != 39) {
    c = (c == '\\') ? spechar(): c;
    k = (k & 255) * 256 + (c & 255);
  }
  val[0] = k;
  return (1);
}

int qstr(int *val)
{
  char c;

  if (!match (quote))
    return (0);
  val[0] = litptr;
  while (ch () != '"') {
    if (ch () == 0)
      break;
    if (litptr >= LITMAX) {
      error ("string space exhausted");
      while (!match (quote))
	if (gch () == 0)
	  break;
      return (1);
    }
    c = gch();
    litq[litptr++] = (c == '\\') ? spechar(): c;
  }
  gch ();
  litq[litptr++] = 0;
  return (1);
}

/*
 *	decode special characters (preceeded by back slashes)
 */
int spechar()
{
  char c;
  c = ch();

  if (c == 'n') c = EOL;
  else if (c == 't') c = TAB;
  else if (c == 'r') c = CR;
  else if (c == 'f') c = FFEED;
  else if (c == 'b') c = BKSP;
  else if (c == '0') c = EOS;
  else if (c == EOS) return(0);

  gch();
  return (c);
}

/*
 *	perform a function call
 *
 *	called from "heir11", this routine will either call the named
 *	function, or if the supplied ptr is zero, will call the contents
 *	of HL
 *
 */
void callfunction(char *ptr)
{
  int nargs;

  nargs = 0;
  blanks ();
  if (ptr == 0)
    gpush ();
  while (!streq (line + lptr, ")")) {
    if (endst ())
      break;
    expression (NO);
    if (ptr == 0)
      swapstk ();
    gpush ();
    nargs = nargs + intsize();
    if (!match (","))
      break;
  }
  needbrack (")");
  if (aflag)
    gnargs(nargs / intsize());
  if (ptr)
    gcall (ptr);
  else
    callstk ();
  stkp = modstk (stkp + nargs);
}

void needlval ()
{
  error ("must be lvalue");
}
